from mwedittypes.tree_differ import get_diff
from mwedittypes.node_differ import get_diff_count


class EditTypes:

    def __init__(self, prev_wikitext='', curr_wikitext='', lang='en', timeout=5, debug=False):
        self.prev_wikitext = prev_wikitext
        self.curr_wikitext = curr_wikitext
        self.lang = lang
        self.timeout = timeout
        self.debug = debug
        self.tree_diff = None
        self.actions = None

    def get_diff(self):
        self.tree_diff = get_diff(self.prev_wikitext, self.curr_wikitext, lang=self.lang,
                                  timeout=self.timeout, debug=self.debug)
        self.actions = get_diff_count(self.tree_diff, lang=self.lang)
        return self.actions



import re
# https://en.wikipedia.org/wiki/Help:Wikitext
# bold, italic, strikethrough, underline
# horizontal rule and line break (hr is not really text formatting but close enough)
# pre and nowiki are for mono-spaced text (I leave out `code` because it generally contains code not text)
# small / big / sub / sup all affect text size
TEXT_FORMATTING_TAGS = ('b', 'i', 's', 'u', 'del', 'ins','hr', 'br','pre', 'nowiki','small',
                         'big', 'sub', 'sup', 'font', 'blockquote', 'span', 'center')
TABLE_ELEMENTS_TAGS = ('th', 'tr', 'td')
LIST_TAGS = ('li', 'dt', 'dd', 'ul', 'ol', 'dl')

# See: https://www.mediawiki.org/wiki/User:TJones_(WMF)/Notes/Spaceless_Writing_Systems_and_Wiki-Projects
NON_WHITESPACE_LANGUAGES = ('bo', 'dz', 'gan', 'ja', 'km', 'lo', 'my', 'th', 'wuu', 'zh', 'zh-classical', 'zh-yue',  # space-less
                            'bug', 'cdo', 'cr', 'hak',  # mixed space/spaceless. could also add: 'jv', 'zh-min-nan'
                            'vi')  # spaces but between syllables and not words

# CJK period/question/exclamation; Bengali full-stops
NON_ENGLISH_PUNCTUATIONS = '。？！।॥'
# This regex identifies end-of-sentence punctuation and new-lines as sentence breaks
# It avoids matching dots between two digits but takes into account ellipses and fullstops.
# fuller explanation:
# [!?...] - 1 or more !, ?, new-line, or non-english stops
# | - or
# (?<!\.) - next character must not precede a fullstop
# \. - next character
# (?<=\d.) - must be preceded by a digit containing a fullstop
# (?!(?<=\d.)\d) -  next character must not be followed by matching dots between two digits
# (?!\.) - next character must not be followed by fullstop
SENTENCE_BREAKS_REGEX = r'[!?\n{0}]+|(?<!\.)\.(?!(?<=\d.)\d)(?!\.)'.format(NON_ENGLISH_PUNCTUATIONS)

NON_ENGLISH_UNICODE = '''[\u0609\u060a\u060c\u060d\u061b\u061e\u061f\u066a\u066b\u066c
\u070a\u070b\u070c\u070d\u07f7\u07f8\u07f9\u0830\u0831\u0832\u0833\u0834\u0835
\u0836\u0837\u0838\u0839\u083a\u083b\u083c\u083d\u083e\u085e\u0964
\u0965\u0970\u09fd\u0a76\u0af0\u0c77\u0c84\u0df4
\u0e4f\u0e5a\u0e5b\u0f04\u0f05\u0f06\u0f07\u0f08\u0f09\u0f0a\u0f0b
\u0f0c\u0f0d\u0f0e\u0f0f\u0f10\u0f11\u0f12\u0f14\u0f85\u0fd0\u0fd1
\u0fd2\u0fd3\u0fd4\u0fd9\u0fda\u104a\u104b\u104c\u104d\u104e\u104f
\u10fb\u1360\u1361\u1362\u1363\u1364\u1365\u1366\u1367\u1368\u166e
\u16eb\u16ec\u16ed\u1735\u1736\u17d4\u17d5\u17d6\u17d8\u17d9\u17da
\u1800\u1801\u1802\u1803\u1804\u1805\u1807\u1808\u1809\u180a\u1944
\u1945\u1a1e\u1a1f\u1aa0\u1aa1\u1aa2\u1aa3\u1aa4\u1aa5\u1aa6\u1aa8
\u1aa9\u1aaa\u1aab\u1aac\u1aad\u1b5a\u1b5b\u1b5c\u1b5d\u1b5e\u1b5f
\u1b60\u1bfc\u1bfd\u1bfe\u1bff\u1c3b\u1c3c\u1c3d\u1c3e\u1c3f\u1c7e
\u1c7f\u1cc0\u1cc1\u1cc2\u1cc3\u1cc4\u1cc5\u1cc6\u1cc7\u1cd3\u2016
\u2017\u2020\u2021\u2022\u2023\u2025\u2026\u2027\u2030\u2031
\u2032\u2033\u2034\u2035\u2036\u2037\u2038\u203b\u203c\u203d\u203e
\u2041\u2042\u2043\u2047\u2048\u2049\u204a\u204b\u204c\u204d\u204e
\u204f\u2050\u2051\u2053\u2055\u2056\u2057\u2058\u2059\u205a\u205b
\u205c\u205d\u205e\u2cf9\u2cfa\u2cfb\u2cfc\u2cfe\u2cff
\u3000-\u303f\uff0c\uff01\uff1f\uff1b\uff1a\uff08\uff3b\u3010\u09E4\u09E5]'''.replace('\n','')

ENGLISH_UNICODE = '[\u00b7\u00bf]'

# https://commons.wikimedia.org/wiki/Commons:File_types
IMAGE_EXTENSIONS = ['.jpg', '.png', '.svg', '.gif', '.jpeg', '.tif', '.bmp', '.webp', '.xcf']
VIDEO_EXTENSIONS = ['.ogv', '.webm', '.mpg', '.mpeg']
AUDIO_EXTENSIONS = ['.ogg', '.mp3', '.mid', '.webm', '.flac', '.wav', '.oga']
MEDIA_EXTENSIONS = list(set(IMAGE_EXTENSIONS + VIDEO_EXTENSIONS + AUDIO_EXTENSIONS))

# build regex that checks for all media extensions
EXTEN_REGEX = ('(' + '|'.join([e + r'\b' for e in MEDIA_EXTENSIONS]) + ')').replace('.', r'\.')
# join in the extension regex with one that requiries at least one alphanumeric and/or a few special characters before it
EXTEN_PATTERN = re.compile(fr'([\w ,().-]+){EXTEN_REGEX}', flags=re.UNICODE)

MEDIA_PREFIXES = ['File', 'Image', 'Media']
CAT_PREFIXES = ['Category']

# Source: for each Wikipedia language code (example shown for "ab"), aliases for namespaces -2 and 6 accessed via this API call:
# https://ab.wikipedia.org/w/api.php?action=query&meta=siteinfo&siprop=namespacealiases|namespaces&format=json&formatversion=2
# Last accessed: 21 December 2021
MEDIA_ALIASES = {
    "ab": ["Медиа", "Файл", "Афаил", "Амедиа", "Изображение"],
    "ace": ["Beureukaih", "Gambar", "Alat", "Berkas"],
    "ady": ["Медиа"],
    "af": ["Lêer", "Beeld"],
    "als": ["Medium", "Datei", "Bild"],
    "am": ["ፋይል", "ስዕል"],
    "an": ["Imachen", "Imagen"],
    "ang": ["Ymele", "Biliþ"],
    "ar": ["ميديا", "صورة", "وسائط", "ملف"],
    "arc": ["ܠܦܦܐ", "ܡܝܕܝܐ"],
    "arz": ["ميديا", "صورة", "وسائط", "ملف"],
    "as": ["চিত্ৰ", "चित्र", "চিত্র", "মাধ্যম"],
    "ast": ["Imaxen", "Ficheru", "Imaxe", "Archivu", "Imagen", "Medios"],
    "atj": ["Tipatcimoctakewin", "Natisinahikaniwoc"],
    "av": ["Медиа", "Файл", "Изображение"],
    "ay": ["Medio", "Archivo", "Imagen"],
    "az": ["Mediya", "Şəkil", "Fayl"],
    "azb": ["رسانه", "تصویر", "مدیا", "فایل", "رسانه‌ای"],
    "ba": ["Медиа", "Рәсем", "Файл", "Изображение"],
    "bar": ["Medium", "Datei", "Bild"],
    "bat-smg": ["Vaizdas", "Medėjė", "Abruozdielis"],
    "bcl": ["Medio", "Ladawan"],
    "be": ["Мультымедыя", "Файл", "Выява"],
    "be-x-old": ["Мэдыя", "Файл", "Выява"],
    "bg": ["Медия", "Файл", "Картинка"],
    "bh": ["मीडिया", "चित्र"],
    "bjn": ["Barakas", "Gambar", "Berkas"],
    "bm": ["Média", "Fichier"],
    "bn": ["চিত্র", "মিডিয়া"],
    "bpy": ["ছবি", "মিডিয়া"],
    "br": ["Skeudenn", "Restr"],
    "bs": ["Mediji", "Slika", "Datoteka", "Medija"],
    "bug": ["Gambar", "Berkas"],
    "bxr": ["Файл", "Меди", "Изображение"],
    "ca": ["Fitxer", "Imatge"],
    "cbk-zam": ["Medio", "Archivo", "Imagen"],
    "cdo": ["文件", "媒體", "圖像", "檔案"],
    "ce": ["Хlум", "Медиа", "Сурт", "Файл", "Медйа", "Изображение"],
    "ceb": ["Payl", "Medya", "Imahen"],
    "ch": ["Litratu"],
    "ckb": ["میدیا", "پەڕگە"],
    "co": ["Immagine"],
    "crh": ["Медиа", "Resim", "Файл", "Fayl", "Ресим"],
    "cs": ["Soubor", "Média", "Obrázok"],
    "csb": ["Òbrôzk", "Grafika"],
    "cu": ["Видъ", "Ви́дъ", "Дѣло", "Срѣдьства"],
    "cv": ["Медиа", "Ӳкерчĕк", "Изображение"],
    "cy": ["Delwedd"],
    "da": ["Billede", "Fil"],
    "de": ["Medium", "Datei", "Bild"],
    "din": ["Ciɛl", "Apamduööt"],
    "diq": ["Medya", "Dosya"],
    "dsb": ["Wobraz", "Dataja", "Bild", "Medija"],
    "dty": ["चित्र", "मिडिया"],
    "dv": ["ފައިލު", "މީޑިއާ", "ފައިލް"],
    "el": ["Εικόνα", "Αρχείο", "Μέσο", "Μέσον"],
    "eml": ["Immagine"],
    "eo": ["Dosiero", "Aŭdvidaĵo"],
    "es": ["Medio", "Archivo", "Imagen"],
    "et": ["Pilt", "Fail", "Meedia"],
    "eu": ["Irudi", "Fitxategi"],
    "ext": ["Archivu", "Imagen", "Mediu"],
    "fa": ["رسانه", "تصویر", "مدیا", "پرونده", "رسانه‌ای"],
    "ff": ["Média", "Fichier"],
    "fi": ["Kuva", "Tiedosto"],
    "fiu-vro": ["Pilt", "Meediä"],
    "fo": ["Miðil", "Mynd"],
    "fr": ["Média", "Fichier"],
    "frp": ["Émâge", "Fichiér", "Mèdia"],
    "frr": ["Medium", "Datei", "Bild"],
    "fur": ["Immagine", "Figure"],
    "fy": ["Ofbyld"],
    "ga": ["Íomhá", "Meán"],
    "gag": ["Mediya", "Medya", "Resim", "Dosya", "Dosye"],
    "gan": ["媒体文件", "文件", "文檔", "档案", "媒體", "图像", "圖像", "媒体", "檔案"],
    "gd": ["Faidhle", "Meadhan"],
    "gl": ["Imaxe", "Ficheiro", "Arquivo", "Imagem"],
    "glk": ["رسانه", "تصویر", "پرونده", "فاىل", "رسانه‌ای", "مديا"],
    "gn": ["Medio", "Imagen", "Ta'ãnga"],
    "gom": ["माध्यम", "मिडिया", "फायल"],
    "gor": ["Gambar", "Berkas"],
    "got": ["𐍆𐌴𐌹𐌻𐌰"],
    "gu": ["દ્રશ્ય-શ્રાવ્ય (મિડિયા)", "દ્રશ્ય-શ્રાવ્ય_(મિડિયા)", "ચિત્ર"],
    "gv": ["Coadan", "Meanyn"],
    "hak": ["文件", "媒體", "圖像", "檔案"],
    "haw": ["Kiʻi", "Waihona", "Pāpaho"],
    "he": ["תמונה", "קו", "מדיה", "קובץ"],
    "hi": ["मीडिया", "चित्र"],
    "hif": ["file", "saadhan"],
    "hr": ["Mediji", "DT", "Slika", "F", "Datoteka"],
    "hsb": ["Wobraz", "Dataja", "Bild"],
    "ht": ["Imaj", "Fichye", "Medya"],
    "hu": ["Kép", "Fájl", "Média"],
    "hy": ["Պատկեր", "Մեդիա"],
    "ia": ["Imagine", "Multimedia"],
    "id": ["Gambar", "Berkas"],
    "ig": ["Nká", "Midia", "Usòrò", "Ákwúkwó orünotu", "Ákwúkwó_orünotu"],
    "ii": ["媒体文件", "文件", "档案", "图像", "媒体"],
    "ilo": ["Midia", "Papeles"],
    "inh": ["Медиа", "Файл", "Изображение"],
    "io": ["Imajo", "Arkivo"],
    "is": ["Miðill", "Mynd"],
    "it": ["Immagine"],
    "ja": ["メディア", "ファイル", "画像"],
    "jbo": ["velsku", "datnyvei"],
    "jv": ["Barkas", "Medhia", "Gambar", "Médhia"],
    "ka": ["მედია", "სურათი", "ფაილი"],
    "kaa": ["Swret", "Таспа", "سۋرەت", "Taspa", "Su'wret", "Сурет", "تاسپا"],
    "kab": ["Tugna"],
    "kbd": ["Медиа", "Файл"],
    "kbp": ["Média", "Fichier"],
    "kg": ["Fisye"],
    "kk": ["Swret", "سۋرەت", "Таспа", "Taspa", "Сурет", "تاسپا"],
    "kl": ["Billede", "Fiileq", "Fil"],
    "km": ["ឯកសារ", "រូបភាព", "មេឌា", "មីឌា"],
    "kn": ["ಚಿತ್ರ", "ಮೀಡಿಯ"],
    "ko": ["미디어", "파일", "그림"],
    "koi": ["Медиа", "Файл", "Изображение"],
    "krc": ["Медиа", "Файл", "Изображение"],
    "ks": ["میڈیا", "فَیِل"],
    "ksh": ["Beld", "Meedije", "Medie", "Belld", "Medium", "Datei", "Meedijum", "Bild"],
    "ku": ["میدیا", "پەڕگە", "Medya", "Wêne"],
    "kv": ["Медиа", "Файл", "Изображение"],
    "kw": ["Restren"],
    "ky": ["Медиа", "Файл"],
    "la": ["Imago", "Fasciculus"],
    "lad": ["Dossia", "Medya", "Archivo", "Dosya", "Imagen", "Meddia"],
    "lb": ["Fichier", "Bild"],
    "lbe": ["Медиа", "Сурат", "Изображение"],
    "lez": ["Медиа", "Mediya", "Файл", "Şəkil", "Изображение"],
    "lfn": ["Fix"],
    "li": ["Afbeelding", "Plaetje", "Aafbeilding"],
    "lij": ["Immaggine", "Immagine"],
    "lmo": ["Immagine", "Imàjine", "Archivi"],
    "ln": ["Média", "Fichier"],
    "lo": ["ສື່ອ", "ສື່", "ຮູບ"],
    "lrc": ["رسانه", "تصویر", "رسانه‌ای", "جانیا", "أسگ", "ڤارئسگأر"],
    "lt": ["Vaizdas", "Medija"],
    "ltg": ["Medeja", "Fails"],
    "lv": ["Attēls"],
    "mai": ["मेडिया", "फाइल"],
    "map-bms": ["Barkas", "Medhia", "Gambar", "Médhia"],
    "mdf": ["Медиа", "Няйф", "Изображение"],
    "mg": ["Rakitra", "Sary", "Média"],
    "mhr": ["Медиа", "Файл", "Изображение"],
    "min": ["Gambar", "Berkas"],
    "mk": ["Податотека", "Медија", "Медиум", "Слика"],
    "ml": ["പ്രമാണം", "ചി", "മീഡിയ", "പ്ര", "ചിത്രം"],
    "mn": ["Медиа", "Файл", "Зураг"],
    "mr": ["चित्र", "मिडिया"],
    "mrj": ["Медиа", "Файл", "Изображение"],
    "ms": ["Fail", "Imej"],
    "mt": ["Midja", "Medja", "Stampa"],
    "mwl": ["Multimédia", "Fexeiro", "Ficheiro", "Arquivo", "Imagem"],
    "my": ["ဖိုင်", "မီဒီယာ"],
    "myv": ["Медия", "Артовкс", "Изображение"],
    "mzn": ["رسانه", "تصویر", "مه‌دیا", "مدیا", "پرونده", "رسانه‌ای"],
    "nah": ["Mēdiatl", "Īxiptli", "Imagen"],
    "nap": ["Fiùra", "Immagine"],
    "nds": ["Datei", "Bild"],
    "nds-nl": ["Ofbeelding", "Afbeelding", "Bestaand"],
    "ne": ["मीडिया", "चित्र"],
    "new": ["किपा", "माध्यम"],
    "nl": ["Bestand", "Afbeelding"],
    "nn": ["Fil", "Bilde", "Filpeikar"],
    "no": ["Fil", "Medium", "Bilde"],
    "nov": [],
    "nrm": ["Média", "Fichier"],
    "nso": ["Seswantšho"],
    "nv": ["Eʼelyaaígíí"],
    "oc": ["Imatge", "Fichièr", "Mèdia"],
    "olo": ["Kuva", "Medii", "Failu"],
    "or": ["ମାଧ୍ୟମ", "ଫାଇଲ"],
    "os": ["Ныв", "Медиа", "Файл", "Изображение"],
    "pa": ["ਤਸਵੀਰ", "ਮੀਡੀਆ"],
    "pcd": ["Média", "Fichier"],
    "pdc": ["Medium", "Datei", "Bild", "Feil"],
    "pfl": ["Dadai", "Medium", "Datei", "Bild"],
    "pi": ["मीडिया", "पटिमा"],
    "pl": ["Plik", "Grafika"],
    "pms": ["Figura", "Immagine"],
    "pnb": ["میڈیا", "تصویر", "فائل"],
    "pnt": ["Εικόνα", "Αρχείον", "Εικόναν", "Μέσον"],
    "ps": ["انځور", "رسنۍ", "دوتنه"],
    "pt": ["Multimédia", "Ficheiro", "Arquivo", "Imagem"],
    "qu": ["Midya", "Imagen", "Rikcha"],
    "rm": ["Multimedia", "Datoteca"],
    "rmy": ["Fişier", "Mediya", "Chitro", "Imagine"],
    "ro": ["Fişier", "Imagine", "Fișier"],
    "roa-rup": ["Fişier", "Imagine", "Fișier"],
    "roa-tara": ["Immagine"],
    "ru": ["Медиа", "Файл", "Изображение"],
    "rue": ["Медіа", "Медиа", "Файл", "Изображение", "Зображення"],
    "rw": ["Dosiye", "Itangazamakuru"],
    "sa": ["चित्रम्", "माध्यमम्", "सञ्चिका", "माध्यम", "चित्रं"],
    "sah": ["Миэдьийэ", "Ойуу", "Билэ", "Изображение"],
    "sat": ["ᱨᱮᱫ", "ᱢᱤᱰᱤᱭᱟ"],
    "sc": ["Immàgini"],
    "scn": ["Immagine", "Mmàggini", "Mèdia"],
    "sd": ["عڪس", "ذريعات", "فائل"],
    "se": ["Fiila"],
    "sg": ["Média", "Fichier"],
    "sh": ["Mediji", "Slika", "Медија", "Datoteka", "Medija", "Слика"],
    "si": ["රූපය", "මාධ්‍යය", "ගොනුව"],
    "sk": ["Súbor", "Obrázok", "Médiá"],
    "sl": ["Slika", "Datoteka"],
    "sq": ["Figura", "Skeda"],
    "sr": ["Датотека", "Medij", "Slika", "Медија", "Datoteka", "Медиј", "Medija", "Слика"],
    "srn": ["Afbeelding", "Gefre"],
    "stq": ["Bielde", "Bild"],
    "su": ["Média", "Gambar"],
    "sv": ["Fil", "Bild"],
    "sw": ["Faili", "Picha"],
    "szl": ["Plik", "Grafika"],
    "ta": ["படிமம்", "ஊடகம்"],
    "tcy": ["ಮಾದ್ಯಮೊ", "ಫೈಲ್"],
    "te": ["ఫైలు", "దస్త్రం", "బొమ్మ", "మీడియా"],
    "tet": ["Imajen", "Arquivo", "Imagem"],
    "tg": ["Акс", "Медиа"],
    "th": ["ไฟล์", "สื่อ", "ภาพ"],
    "ti": ["ፋይል", "ሜድያ"],
    "tk": ["Faýl"],
    "tl": ["Midya", "Talaksan"],
    "tpi": ["Fail"],
    "tr": ["Medya", "Resim", "Dosya", "Ortam"],
    "tt": ["Медиа", "Рәсем", "Файл", "Räsem", "Изображение"],
    "ty": ["Média", "Fichier"],
    "tyv": ["Медиа", "Файл", "Изображение"],
    "udm": ["Медиа", "Файл", "Суред", "Изображение"],
    "ug": ["ۋاسىتە", "ھۆججەت"],
    "uk": ["Медіа", "Медиа", "Файл", "Изображение", "Зображення"],
    "ur": ["میڈیا", "تصویر", "وسیط", "زریعہ", "فائل", "ملف"],
    "uz": ["Mediya", "Tasvir", "Fayl"],
    "vec": ["Immagine", "Imàjine", "Mèdia"],
    "vep": ["Pilt", "Fail"],
    "vi": ["Phương_tiện", "Tập_tin", "Hình", "Tập tin", "Phương tiện"],
    "vls": ["Afbeelding", "Ofbeeldienge"],
    "vo": ["Ragiv", "Magod", "Nünamakanäd"],
    "wa": ["Imådje"],
    "war": ["Medya", "Fayl", "Paypay"],
    "wo": ["Xibaarukaay", "Dencukaay"],
    "wuu": ["文件", "档案", "图像", "媒体"],
    "xal": ["Аһар", "Боомг", "Изображение", "Зург"],
    "xmf": ["მედია", "სურათი", "ფაილი"],
    "yi": ["מעדיע", "תמונה", "טעקע", "בילד"],
    "yo": ["Fáìlì", "Amóhùnmáwòrán", "Àwòrán"],
    "za": ["媒体文件", "文件", "档案", "图像", "媒体"],
    "zea": ["Afbeelding", "Plaetje"],
    "zh": ["媒体文件", "F", "文件", "媒體", "档案", "图像", "圖像", "媒体", "檔案"],
    "zh-classical": ["文件", "媒體", "圖像", "檔案"],
    "zh-min-nan": ["tóng-àn", "文件", "媒體", "Mûi-thé", "圖像", "檔案"],
    "zh-yue": ["檔", "档", "文件", "图", "媒體", "圖", "档案", "图像", "圖像", "媒体", "檔案"],
}

# Source: for each Wikipedia language code (example shown for "ab"), aliases for namespace 14 accessed via this API call:
# https://ab.wikipedia.org/w/api.php?action=query&meta=siteinfo&siprop=namespacealiases|namespaces&format=json&formatversion=2
# Last accessed: 21 December 2021
CAT_ALIASES = {
    "ab": ["Категория", "Акатегориа"],
    "ace": ["Kawan", "Kategori"],
    "af": ["Kategorie"],
    "ak": ["Nkyekyem"],
    "als": ["Kategorie"],
    "am": ["መደብ"],
    "an": ["Categoría"],
    "ang": ["Flocc"],
    "ar": ["تصنيف"],
    "arc": ["ܣܕܪܐ"],
    "arz": ["تصنيف"],
    "as": ["CAT", "শ্ৰেণী", "श्रेणी", "শ্রেণী"],
    "ast": ["Categoría"],
    "atj": ["Tipanictawin"],
    "av": ["Категория"],
    "ay": ["Categoría"],
    "az": ["Kateqoriya"],
    "azb": ["بؤلمه"],
    "ba": ["Төркөм", "Категория"],
    "bar": ["Kategorie"],
    "bat-smg": ["Kategorija", "Kateguorėjė"],
    "bcl": ["Kategorya"],
    "be": ["Катэгорыя"],
    "be-x-old": ["Катэгорыя"],
    "bg": ["Категория"],
    "bh": ["श्रेणी"],
    "bjn": ["Tumbung", "Kategori"],
    "bm": ["Catégorie"],
    "bn": ["বিষয়শ্রেণী", "വിഭാഗം"],
    "bpy": ["থাক"],
    "br": ["Rummad"],
    "bs": ["Kategorija"],
    "bug": ["Kategori"],
    "bxr": ["Категори", "Категория"],
    "ca": ["Categoria"],
    "cbk-zam": ["Categoría"],
    "cdo": ["分類"],
    "ce": ["Категори", "Тоба", "Кадегар"],
    "ceb": ["Kategoriya"],
    "ch": ["Katigoria"],
    "ckb": ["پ", "پۆل"],
    "co": ["Categoria"],
    "crh": ["Категория", "Kategoriya"],
    "cs": ["Kategorie"],
    "csb": ["Kategòrëjô"],
    "cu": ["Катигорї", "Категория", "Катигорїꙗ"],
    "cv": ["Категори"],
    "cy": ["Categori"],
    "da": ["Kategori"],
    "de": ["Kategorie"],
    "din": ["Bekätakthook"],
    "diq": ["Kategoriye", "Kategori"],
    "dsb": ["Kategorija"],
    "dty": ["श्रेणी"],
    "dv": ["ޤިސްމު"],
    "el": ["Κατηγορία"],
    "eml": ["Categoria"],
    "eo": ["Kategorio"],
    "es": ["CAT", "Categoría"],
    "et": ["Kategooria"],
    "eu": ["Kategoria"],
    "ext": ["Categoría", "Categoria"],
    "fa": ["رده"],
    "ff": ["Catégorie"],
    "fi": ["Luokka"],
    "fiu-vro": ["Katõgooria"],
    "fo": ["Bólkur"],
    "fr": ["Catégorie"],
    "frp": ["Catègorie"],
    "frr": ["Kategorie"],
    "fur": ["Categorie"],
    "fy": ["Kategory"],
    "ga": ["Rang", "Catagóir"],
    "gag": ["Kategori", "Kategoriya"],
    "gan": ["分類", "分类"],
    "gd": ["Roinn-seòrsa"],
    "gl": ["Categoría"],
    "glk": ["جرگه", "رده"],
    "gn": ["Ñemohenda"],
    "gom": ["वर्ग", "श्रेणी"],
    "gor": ["Dalala"],
    "got": ["𐌷𐌰𐌽𐍃𐌰"],
    "gu": ["શ્રેણી", "CAT", "શ્રે"],
    "gv": ["Ronney"],
    "hak": ["分類"],
    "haw": ["Māhele"],
    "he": ["קטגוריה", "קט"],
    "hi": ["श्र", "श्रेणी"],
    "hif": ["vibhag"],
    "hr": ["CT", "KT", "Kategorija"],
    "hsb": ["Kategorija"],
    "ht": ["Kategori"],
    "hu": ["Kategória"],
    "hy": ["Կատեգորիա"],
    "ia": ["Categoria"],
    "id": ["Kategori"],
    "ie": ["Categorie"],
    "ig": ["Ébéonọr", "Òtù"],
    "ii": ["分类"],
    "ilo": ["Kategoria"],
    "inh": ["ОагӀат"],
    "io": ["Kategorio"],
    "is": ["Flokkur"],
    "it": ["CAT", "Categoria"],
    "ja": ["カテゴリ"],
    "jbo": ["klesi"],
    "jv": ["Kategori"],
    "ka": ["კატეგორია"],
    "kaa": ["Sanat", "Kategoriya", "Санат", "سانات"],
    "kab": ["Taggayt"],
    "kbd": ["Категория", "Категориэ"],
    "kbp": ["Catégorie"],
    "kg": ["Kalasi"],
    "kk": ["Sanat", "Санат", "سانات"],
    "kl": ["Sumut_atassuseq", "Kategori", "Sumut atassuseq"],
    "km": ["ចំនាត់ថ្នាក់ក្រុម", "ចំណាត់ក្រុម", "ចំណាត់ថ្នាក់ក្រុម"],
    "kn": ["ವರ್ಗ"],
    "ko": ["분류"],
    "koi": ["Категория"],
    "krc": ["Категория"],
    "ks": ["زٲژ"],
    "ksh": ["Saachjropp", "Saachjrop", "Katejori", "Kategorie", "Saachjrupp", "Kattejori", "Sachjrop"],
    "ku": ["Kategorî", "پۆل"],
    "kv": ["Категория"],
    "kw": ["Class", "Klass"],
    "ky": ["Категория"],
    "la": ["Categoria"],
    "lad": ["Kateggoría", "Katēggoría", "Categoría"],
    "lb": ["Kategorie"],
    "lbe": ["Категория"],
    "lez": ["Категория"],
    "lfn": ["Categoria"],
    "li": ["Categorie", "Kategorie"],
    "lij": ["Categorîa", "Categoria"],
    "lmo": ["Categuria", "Categoria"],
    "ln": ["Catégorie"],
    "lo": ["ໝວດ"],
    "lrc": ["دأسە"],
    "lt": ["Kategorija"],
    "ltg": ["Kategoreja"],
    "lv": ["Kategorija"],
    "mai": ["CA", "श्रेणी"],
    "map-bms": ["Kategori"],
    "mdf": ["Категорие", "Категория"],
    "mg": ["Sokajy", "Catégorie"],
    "mhr": ["Категория", "Категорий"],
    "min": ["Kategori"],
    "mk": ["Категорија"],
    "ml": ["വിഭാഗം", "വി", "വർഗ്ഗം", "വ"],
    "mn": ["Ангилал"],
    "mr": ["वर्ग"],
    "mrj": ["Категори", "Категория"],
    "ms": ["Kategori"],
    "mt": ["Kategorija"],
    "mwl": ["Catadorie", "Categoria"],
    "my": ["ကဏ္ဍ"],
    "myv": ["Категория"],
    "mzn": ["رج", "رده"],
    "nah": ["Neneuhcāyōtl", "Categoría"],
    "nap": ["Categurìa", "Categoria"],
    "nds": ["Kategorie"],
    "nds-nl": ["Categorie", "Kattegerie", "Kategorie"],
    "ne": ["श्रेणी"],
    "new": ["पुचः"],
    "nl": ["Categorie"],
    "nn": ["Kategori"],
    "no": ["Kategori"],
    "nrm": ["Catégorie"],
    "nso": ["Setensele"],
    "nv": ["Tʼááłáhági_átʼéego", "Tʼááłáhági átʼéego"],
    "oc": ["Categoria"],
    "olo": ["Kategourii"],
    "or": ["ବିଭାଗ", "ଶ୍ରେଣୀ"],
    "os": ["Категори"],
    "pa": ["ਸ਼੍ਰੇਣੀ"],
    "pcd": ["Catégorie"],
    "pdc": ["Abdeeling", "Kategorie"],
    "pfl": ["Kadegorie", "Sachgrubb", "Kategorie"],
    "pi": ["विभाग"],
    "pl": ["Kategoria"],
    "pms": ["Categorìa"],
    "pnb": ["گٹھ"],
    "pnt": ["Κατηγορίαν"],
    "ps": ["وېشنيزه"],
    "pt": ["Categoria"],
    "qu": ["Katiguriya"],
    "rm": ["Categoria"],
    "rmy": ["Shopni"],
    "ro": ["Categorie"],
    "roa-rup": ["Categorie"],
    "roa-tara": ["Categoria"],
    "ru": ["Категория", "К"],
    "rue": ["Категория", "Катеґорія"],
    "rw": ["Ikiciro"],
    "sa": ["वर्गः"],
    "sah": ["Категория"],
    "sat": ["ᱛᱷᱚᱠ"],
    "sc": ["Categoria"],
    "scn": ["Catigurìa"],
    "sd": ["زمرو"],
    "se": ["Kategoriija"],
    "sg": ["Catégorie"],
    "sh": ["Kategorija", "Категорија"],
    "si": ["ප්‍රවර්ගය"],
    "sk": ["Kategória"],
    "sl": ["Kategorija"],
    "sq": ["Kategoria", "Kategori"],
    "sr": ["Kategorija", "Категорија"],
    "srn": ["Categorie", "Guru"],
    "stq": ["Kategorie"],
    "su": ["Kategori"],
    "sv": ["Kategori"],
    "sw": ["Jamii"],
    "szl": ["Kategoryjo", "Kategoria"],
    "ta": ["பகுப்பு"],
    "tcy": ["ವರ್ಗೊ"],
    "te": ["వర్గం"],
    "tet": ["Kategoría", "Kategoria"],
    "tg": ["Гурӯҳ"],
    "th": ["หมวดหมู่"],
    "ti": ["መደብ"],
    "tk": ["Kategoriýa"],
    "tl": ["Kategorya", "Kaurian"],
    "tpi": ["Grup"],
    "tr": ["Kategori", "KAT"],
    "tt": ["Төркем", "Törkem", "Категория"],
    "ty": ["Catégorie"],
    "tyv": ["Аңгылал", "Категория"],
    "udm": ["Категория"],
    "ug": ["تۈر"],
    "uk": ["Категория", "Категорія"],
    "ur": ["زمرہ"],
    "uz": ["Turkum", "Kategoriya"],
    "vec": ["Categoria"],
    "vep": ["Kategorii"],
    "vi": ["Thể_loại", "Thể loại"],
    "vls": ["Categorie"],
    "vo": ["Klad"],
    "wa": ["Categoreye"],
    "war": ["Kaarangay"],
    "wo": ["Wàll", "Catégorie"],
    "wuu": ["分类"],
    "xal": ["Янз", "Әәшл"],
    "xmf": ["კატეგორია"],
    "yi": ["קאטעגאריע", "קאַטעגאָריע"],
    "yo": ["Ẹ̀ka"],
    "za": ["分类"],
    "zea": ["Categorie"],
    "zh": ["分类", "分類", "CAT"],
    "zh-classical": ["分類", "CAT"],
    "zh-min-nan": ["分類", "Lūi-pia̍t"],
    "zh-yue": ["分类", "分類", "类", "類"],
}



import time

from anytree import PostOrderIter, NodeMixin
from anytree.util import leftsibling
import mwparserfromhell as mw

from mwedittypes.constants import *


# equivalent of main function
def get_diff(prev_wikitext, curr_wikitext, lang='en', timeout=2, debug=False):
    """Run through full process of getting tree diff between two wikitext revisions."""
    # To provide proper structure, need all content to be nested under a section
    prev_tree = WikitextTree(wikitext=prev_wikitext, lang=lang)
    curr_tree = WikitextTree(wikitext=curr_wikitext, lang=lang)
    d = Differ(prev_tree, curr_tree, timeout=timeout)
    diff = d.get_corresponding_nodes(debug=debug)
    result = diff.post_process(prev_tree.secname_to_text, curr_tree.secname_to_text, lang=lang)
    # this helps the node differ know that the lede is also a new section
    # otherwise depends on headings to track changes to sections
    if not prev_wikitext:
        result['prev-no-content'] = True
    return result


# helper functions for handling mwparserfromhell / wikitext
def simple_node_class(mwnode, lang='en'):
    """e.g., "<class 'mwparserfromhell.nodes.heading.Heading'>" -> "Heading"."""
    if type(mwnode) == str:
        return 'Text'
    else:
        nc = str(type(mwnode)).split('.')[-1].split("'")[0]
        if nc == 'Wikilink':
            n_prefix = mwnode.title.split(':', maxsplit=1)[0].lower()
            if n_prefix in [m.lower() for m in MEDIA_PREFIXES + MEDIA_ALIASES.get(lang, [])]:
                nc = 'Media'
            elif n_prefix in [c.lower() for c in CAT_PREFIXES + CAT_ALIASES.get(lang, [])]:
                nc = 'Category'
        elif nc == 'Tag':
            tag_type = str(mwnode.tag)
            if tag_type in TEXT_FORMATTING_TAGS:
                return 'Text Formatting'
            elif tag_type in LIST_TAGS:
                return 'List'
            elif tag_type == 'table':
                return 'Table'
            elif tag_type in TABLE_ELEMENTS_TAGS:
                return 'Table Element'
            elif tag_type == 'gallery':
                return 'Gallery'
            elif tag_type == 'ref':
                return 'Reference'
            elif tag_type == 'noinclude':
                return 'Comment'
            # any others I missed -- e.g., div, meta, etc.
            else:
                return tag_type.capitalize() + ' Tag'
        return nc


def sec_to_name(mwsection, sidx):
    """Converts a section to an interpretible and unique name."""
    if sidx == 0:
        sectitle = 'Lede'
    else:
        try:
            sectitle = str(mwsection.nodes[0])
        except IndexError:
            sectitle = ''
    return f'{sidx}: {sectitle}'


def node_to_name(mwnode, lang='en'):
    """Converts a mwparserfromhell node to an interpretible name."""
    n_txt = mwnode.replace("\n", "\\n")
    if len(n_txt) > 13:
        return f'{simple_node_class(mwnode, lang)}: {n_txt[:10]}...'
    else:
        return f'{simple_node_class(mwnode, lang)}: {n_txt}'


def extract_text(mwnode, lang='en'):
    """Extract what text would be displayed from any node."""
    ntype = simple_node_class(mwnode, lang)
    if ntype == 'Text':
        return str(mwnode)
    elif ntype == 'HTMLEntity':
        return mwnode.normalize()
    elif ntype == 'Wikilink':
        if mwnode.text:
            return mwnode.text.strip_code()
        else:
            return mwnode.title.strip_code()
    elif ntype == 'ExternalLink' and mwnode.title:
        return mwnode.title.strip_code()
    # tables can have tons of nested references etc. so can't just go through standard strip_code
    elif ntype == 'Table':
        # don't collapse whitespace for tables because otherwise strip_code sometimes merges text across cells
        return mwnode.contents.strip_code(collapse=False)
    # divs/galleries almost never have true text content and can be super messy so best ignored when building Text
    elif ntype == 'Text Formatting':
        return ''.join(extract_text(mwn) for mwn in mwnode.contents.nodes)
    # Heading, Template, Comment, Argument, Category, Media, References, URLs without display text
    # Tags not listed here (div, gallery, etc.) that almost never have true text content and can be super messy
    # Table elements (they duplicate the text if included)
    else:
        return ''


def find_nested_media(wikitext, max_link_length=240):
    """Case-insensitive search for media files (lacking brackets) in wikitext -- i.e. in Templates and Galleries.

    For setting max_link_length: https://commons.wikimedia.org/wiki/Commons:File_naming#Length
    """
    lc_wt = wikitext.lower()
    media = []
    end = 0
    while True:
        m = EXTEN_PATTERN.search(lc_wt, pos=end)
        if m is None:
            break
        start, end = m.span()
        if end - start <= max_link_length:
            media.append(wikitext[start:end].strip())
    return media


class OrderedNode(NodeMixin):
    """
    Extension of anytree library node to support tree differ.
    """

    def __init__(self, name, ntype='Text', text_hash=None, idx=-1, text='', char_offset=-1, section=None,
                 parent=None, children=None):
        super(OrderedNode, self).__init__()
        self.name = name  # For debugging purposes
        self.ntype = ntype  # Different node types can be treated differently when computing equality
        self.text = str(text)  # Text that can then be passed to a diffing library
        # Used for quickly computing equality for most nodes.
        # Generally this just a simple hash of self.text (wikitext associated with a node) but
        # the text hash for sections and paragraphs is based on all the content within the section/paragraph
        # so it can be used for pruning while self.text is just the text that creates the section/paragraph
        # e.g., "==Section==\nThis is a section." would have as text "==Section==" but hash the full.
        # so the Differ doesn't identify a section/paragraph as changing when content within it is changed
        if text_hash is None:
            self.text_hash = hash(self.text)
        else:
            self.text_hash = hash(str(text_hash))
        self.idx = idx  # Used by Differ -- Post order on tree from 0...# nodes - 1
        self.char_offset = char_offset  # make it easy to find node in section text
        self.section = section  # section that the node is a part of -- useful for formatting final diff
        self.parent = parent
        if children:
            self.children = children

    def leftmost(self):
        return self.idx if self.is_leaf else self.children[0].leftmost()

    def unnest(self, lang='en'):
        """Build tree of document nodes by recursing within a single wikitext node.

        This approach starts with a single wikitext node -- e.g., a single Tag node with nested link nodes etc.:
        <ref>{{cite web|title=[[Gallery]]|url=http://digital.belvedere.at|publisher=Digitales Belvedere}}</ref>
        and splits it into its component pieces to then identify what has changed between revisions.

        Example above would take a Reference node as input and build the following tree (in-place):
        <--rest-of-tree-- Reference <--child-of-- Template (cite web) <--child-of-- WikiLink (Gallery)
                                                                ^--------child-of-- External Link (http://digital...)
        """
        if self.ntype == 'Gallery':
            # strip leading / trailing gallery tags so parser correctly parses everything in between
            # otherwise links, formatting, etc. is treated as text
            gallery_start = self.text.find('>')
            gallery_end = self.text.rfind('<')
            try:
                # the break is a hack; it'll be skipped in the ifilter loop; otherwise first image skipped
                wt = mw.parse('<br>' + self.text[gallery_start+1:gallery_end])
            except Exception:  # fallback
                wt = mw.parse(self.text)
        else:
            wt = mw.parse(self.text)
        parent_node = self
        base_offset = self.char_offset
        parent_ranges = [(0, len(self.text), self)]  # (start idx of node, end idx of node, node object)
        for idx, nn in enumerate(wt.ifilter(recursive=True)):
            if idx == 0:
                continue  # skip root node -- already set or placeholder <br> node for galleries
            ntype = simple_node_class(nn, lang)
            if ntype == 'Text':
                # media w/o bracket will be IDed as text by mwparserfromhell
                # templates / galleries are where we find this nested media
                if self.ntype == 'Template' or self.ntype == 'Gallery':
                    media = find_nested_media(str(nn))
                    for m in media:
                        nn_node = OrderedNode(f'Media: {m[:10]}...',
                                              ntype='Media',
                                              text=m,
                                              char_offset=base_offset + self.text.find(str(m), parent_ranges[0][0]),
                                              section=self.section,
                                              parent=self)
            # tables are very highly-structured and produce a ton of nodes (each cell and more)
            # so we just extract links, formatting, etc. that appears in the table and skip the cells
            # because changes to those will generally be caught in the overall table changes and text changes
            # this leads to much faster parsing in exchange for not knowing how many table cells were edited
            elif ntype == 'Table Element':
                pass
            else:
                node_start = self.text.find(str(nn), parent_ranges[0][0])  # start looking from the start of the latest node
                # identify direct parent of node
                for parent in parent_ranges:
                    if node_start < parent[1]:  # falls within parent range
                        parent_node = parent[2]
                        break
                nn_node = OrderedNode(node_to_name(nn, lang=lang), ntype=ntype, text=nn,
                                      char_offset=base_offset + node_start,
                                      section=self.section, parent=parent_node)
                parent_ranges.insert(0, (node_start, node_start + len(nn), nn_node))

    def dump(self, debug=False):
        result = {'type': self.ntype,
                  'text': self.text,
                  'section': self.section}
        if debug:
            result['name'] = self.name
            result['offset'] = self.char_offset
        return result


class WikitextTree:
    """
    Tree structure for wikitext based on mwparserfromhell
    """

    def __init__(self, wikitext, lang="en"):
        self.lang = lang
        self.root = OrderedNode('root', ntype="Article")
        self.secname_to_text = {}
        self.wikitext_to_tree(wikitext)

    def wikitext_to_tree(self, wikitext):
        """Build tree of document nodes from Wikipedia article.

        This approach builds a tree with an artificial 'root' node on the 1st level,
        all of the article sections nested flatly underneath (including the Lede section),
        and all of the text, link, template, etc. nodes nested under their respective sections.
        """
        wt = mw.parse(wikitext)
        for sidx, s in enumerate(wt.get_sections(flat=True)):
            if s:
                sec_hash = sec_to_name(s, sidx)
                sec_text = ''.join([str(n) for n in s.nodes])
                self.secname_to_text[sec_hash] = sec_text
                s_node = OrderedNode(sec_hash, ntype="Section", text=s.nodes[0], text_hash=sec_text, char_offset=0,
                                     section=sec_hash, parent=self.root)
                char_offset = 0
                for n in s.nodes:
                    n_node = OrderedNode(node_to_name(n, self.lang), ntype=simple_node_class(n, self.lang), text=n,
                                         char_offset=char_offset, section=s_node.name, parent=s_node)
                    char_offset += len(str(n))
            # empty lede
            else:
                sec_hash = sec_to_name(s, sidx)
                sec_text = ''
                self.secname_to_text[sec_hash] = sec_text
                s_node = OrderedNode(sec_hash, ntype="Section", text=sec_text, text_hash=sec_text, char_offset=0,
                                     section=sec_hash, parent=self.root)

    def expand_nested(self):
        """Expand nested nodes in tree -- e.g., Ref tags with templates/links contained in them."""
        for n in PostOrderIter(self.root):
            if n.ntype not in ('Article', 'Section', 'Heading', 'Text'):  # leaves tag, link, etc.
                n.unnest(self.lang)


class Differ:
    """
    Find structural differences between two WikitextTrees
    """

    def __init__(self, t1, t2, timeout=2, expand_nodes=True):
        self.prune_trees(t1, t2, expand_nodes)
        self.t1 = []
        self.t2 = []
        for i, n in enumerate(PostOrderIter(t1.root)):
            n.idx = i
            self.t1.append(n)
        for i, n in enumerate(PostOrderIter(t2.root)):
            n.idx = i
            self.t2.append(n)
        self.timeout = time.time() + timeout
        self.ins_cost = 1
        self.rem_cost = 1
        self.chg_cost = 1
        self.nodetype_chg_cost = 10  # arbitrarily high to encourage remove+insert when node types change

        # Permanent store of transactions such that transactions[x][y] is the minimum
        # transactions to get from the sub-tree rooted at node x (in tree1) to the sub-tree
        # rooted at node y (in tree2).
        self.transactions = {None: {}}
        # Indices for each transaction, to avoid high performance cost of creating the
        # transactions multiple times
        self.transaction_to_idx = {None: {None: 0}}
        # All possible transactions
        self.idx_to_transaction = [(None, None)]

        idx_transaction = 1  # starts with nulls inserted

        transactions = {None: {None: []}}

        # Populate transaction stores
        for i in range(0, len(self.t1)):
            transactions[i] = {None: []}
            self.transaction_to_idx[i] = {None: idx_transaction}
            idx_transaction += 1
            self.idx_to_transaction.append((i, None))
            for j in range(0, len(self.t2)):
                transactions[None][j] = []
                transactions[i][j] = []
                self.transaction_to_idx[None][j] = idx_transaction
                idx_transaction += 1
                self.idx_to_transaction.append((None, j))
                self.transaction_to_idx[i][j] = idx_transaction
                idx_transaction += 1
                self.idx_to_transaction.append((i, j))
            self.transactions[i] = {}
        self.populate_transactions(transactions)

    def prune_trees(self, t1, t2, expand_nodes=False):
        """Quick heuristic preprocessing to reduce tree differ time by removing matching sections."""
        self.prune_sections(t1, t2)
        # more than 500 nodes altogether even after pruning and before unnesting -- just diff sections
        if (sum(1 for n in PostOrderIter(t1.root)) + sum(1 for n in PostOrderIter(t2.root))) > 500:
            self.prune_to_sections(t1, t2)
        # seems like manageable number of total nodes -- unnest fully before diffing
        elif expand_nodes and (sum([len(mw.parse(n.text).filter()) for n in PostOrderIter(t1.root)]) +
                               sum([len(mw.parse(n.text).filter()) for n in PostOrderIter(t2.root)])) < 1000:
            t1.expand_nested()
            t2.expand_nested()

    def prune_to_sections(self, t1, t2):
        """Remove all non-section nodes."""
        for n in PostOrderIter(t1.root):
            if n.ntype == 'Section':
                n.children = []
        for n in PostOrderIter(t2.root):
            if n.ntype == 'Section':
                n.children = []

    def prune_sections(self, t1, t2):
        """Prune nodes from any sections that align across revisions"""
        t1_sections = [n for n in PostOrderIter(t1.root) if n.ntype == "Section"]
        t2_sections = [n for n in PostOrderIter(t2.root) if n.ntype == "Section"]
        for secnode1 in t1_sections:
            for sn2_idx in range(len(t2_sections)):
                secnode2 = t2_sections[sn2_idx]
                if secnode1.text_hash == secnode2.text_hash:
                    # assumes sections aren't hierarchical in tree
                    # or if they are, the text_hash must also include nested sections
                    secnode1.children = []
                    secnode2.children = []
                    t2_sections.pop(sn2_idx)  # only match once
                    break

    def get_key_roots(self, tree):
        """Get keyroots (node has a left sibling or is the root) of a tree"""
        for on in tree:
            if on.is_root or leftsibling(on) is not None:
                yield on

    def populate_transactions(self, transactions):
        """Populate self.transactions with minimum transactions between all possible trees"""
        for kr1 in self.get_key_roots(self.t1):
            # Make transactions for tree -> null
            i_nulls = []
            for ii in range(kr1.leftmost(), kr1.idx + 1):
                i_nulls.append(self.transaction_to_idx[ii][None])
                transactions[ii][None] = i_nulls.copy()
            for kr2 in self.get_key_roots(self.t2):
                # Make transactions of null -> tree
                j_nulls = []
                for jj in range(kr2.leftmost(), kr2.idx + 1):
                    j_nulls.append(self.transaction_to_idx[None][jj])
                    transactions[None][jj] = j_nulls.copy()

                # get the diff
                self.find_minimum_transactions(kr1, kr2, transactions)
                if time.time() > self.timeout:
                    self.transactions = None
                    return

        if self.transactions:
            for i in range(0, len(self.t1)):
                for j in range(0, len(self.t2)):
                    if self.transactions.get(i, {}).get(j) and len(self.transactions[i][j]) > 0:
                        self.transactions[i][j] = tuple([self.idx_to_transaction[idx] for idx in self.transactions[i][j]])

    def get_node_distance(self, n1, n2):
        """
        Get the cost of:
        * removing a node from the first tree,
        * inserting a node into the second tree,
        * or relabelling a node from the first tree to a node from the second tree.
        """
        if n1 is None and n2 is None:
            return 0
        elif n1 is None:
            return self.ins_cost
        elif n2 is None:
            return self.rem_cost
        # Inserts/Removes are easy. Changes are more complicated and should only be within same node type.
        # Use arbitrarily high-value for nodetype changes to effectively ban.
        elif n1.ntype != n2.ntype:
            return self.nodetype_chg_cost
        elif n1.text_hash == n2.text_hash:
            return 0
        # if both sections, assert no cost (changes will be captured by headings)
        # except we want to detect moves of sections
        elif n1.ntype == 'Section':
            if not n1.children and not n2.children:
                return self.chg_cost
            return 0
        # otherwise, same node types and not the same, then change cost
        else:
            return self.chg_cost

    def get_lowest_cost(self, rc, ic, cc):
        min_cost = rc
        index = 0
        if ic < min_cost:
            index = 1
            min_cost = ic
        if cc < min_cost:
            index = 2
        return index

    def find_minimum_transactions(self, kr1, kr2, transactions):
        """Find the minimum transactions to get from the first tree to the second tree."""
        for i in range(kr1.leftmost(), kr1.idx + 1):
            if i == kr1.leftmost():
                i_minus_1 = None
            else:
                i_minus_1 = i - 1
            n1 = self.t1[i]
            for j in range(kr2.leftmost(), kr2.idx + 1):
                if j == kr2.leftmost():
                    j_minus_1 = None
                else:
                    j_minus_1 = j - 1
                n2 = self.t2[j]

                if n1.leftmost() == kr1.leftmost() and n2.leftmost() == kr2.leftmost():
                    rem = transactions[i_minus_1][j]
                    ins = transactions[i][j_minus_1]
                    chg = transactions[i_minus_1][j_minus_1]
                    node_distance = self.get_node_distance(n1, n2)
                    # cost of each transaction
                    transaction = self.get_lowest_cost(len(rem) + self.rem_cost,
                                                       len(ins) + self.ins_cost,
                                                       len(chg) + node_distance)
                    if transaction == 0:
                        # record a remove
                        rc = rem.copy()
                        rc.append(self.transaction_to_idx[i][None])
                        transactions[i][j] = rc
                    elif transaction == 1:
                        # record an insert
                        ic = ins.copy()
                        ic.append(self.transaction_to_idx[None][j])
                        transactions[i][j] = ic
                    else:
                        # If nodes i and j are different, record a change, otherwise there is no transaction
                        transactions[i][j] = chg.copy()
                        if node_distance == 1:
                            transactions[i][j].append(self.transaction_to_idx[i][j])

                    self.transactions[i][j] = transactions[i][j].copy()
                else:
                    # Previous transactions, leading up to a remove, insert or change
                    rem = transactions[i_minus_1][j]
                    ins = transactions[i][j_minus_1]

                    if n1.leftmost() - 1 < kr1.leftmost():
                        k1 = None
                    else:
                        k1 = n1.leftmost() - 1
                    if n2.leftmost() - 1 < kr2.leftmost():
                        k2 = None
                    else:
                        k2 = n2.leftmost() - 1
                    chg = transactions[k1][k2]

                    transaction = self.get_lowest_cost(len(rem) + self.rem_cost,
                                                       len(ins) + self.ins_cost,
                                                       len(chg) + len(self.transactions[i][j]))
                    if transaction == 0:
                        # record a remove
                        rc = rem.copy()
                        rc.append(self.transaction_to_idx[i][None])
                        transactions[i][j] = rc
                    elif transaction == 1:
                        # record an insert
                        ic = ins.copy()
                        ic.append(self.transaction_to_idx[None][j])
                        transactions[i][j] = ic
                    else:
                        # record a change
                        cc = chg.copy()
                        cc.extend(self.transactions[i][j])
                        transactions[i][j] = cc

    def get_corresponding_nodes(self, debug=False):
        """Explain transactions.

        Skip 'Section' operations as they aren't real nodes and
        any changes to them will be captured via Headings etc.
        """
        if self.transactions:
            transactions = self.transactions[len(self.t1) - 1][len(self.t2) - 1]
            remove = []
            insert = []
            change = []
            for i in range(0, len(transactions)):
                if transactions[i][0] is None:
                    ins_node = self.t2[transactions[i][1]]
                    # this second clause allows us to later detect moved sections
                    if ins_node.ntype != 'Section' or not ins_node.children:
                        insert.append(ins_node)
                elif transactions[i][1] is None:
                    rem_node = self.t1[transactions[i][0]]
                    if rem_node.ntype != 'Section' or not rem_node.children:
                        remove.append(rem_node)
                else:
                    prev_node = self.t1[transactions[i][0]]
                    curr_node = self.t2[transactions[i][1]]
                    if prev_node.ntype != 'Section' or not prev_node.children:
                        change.append((prev_node, curr_node))
            diff = {'remove': remove, 'insert': insert, 'change': change}
            self.detect_moves(diff)
            return Diff(nodes_removed=diff['remove'], nodes_inserted=diff['insert'],
                        nodes_changed=diff['change'], nodes_moved=diff['move'],
                        debug=debug)
        else:
            return Diff(nodes_removed=[], nodes_inserted=[],
                        nodes_changed=[], nodes_moved=[])

    def detect_moves(self, diff):
        """Detect when nodes were moved (as opposed to removed/inserted/changed) and update diff."""

        # build list of all prev and curr nodes to compare for matches
        prev_nodes = [('remove', i, pn) for i, pn in enumerate(diff['remove'])]
        curr_nodes = [('insert', j, cn) for j, cn in enumerate(diff['insert'])]
        for k in range(len(diff['change'])):
            prev_nodes.append(('change', k, diff['change'][k][0]))
            curr_nodes.append(('change', k, diff['change'][k][1]))

        # loop through prev/curr nodes and look for matches. constraints:
        # * nodes can only match with one other node
        # * if a node is part of a change, make sure it's corresponding node is moved to insert/remove accordingly
        prev_moved = []
        curr_moved = []
        curr_found = set()
        add_to_insert = {}
        add_to_remove = {}
        for pet, pidx, pn in prev_nodes:
            for cet, cidx, cn in curr_nodes:
                cid = f'{cet}-{cidx}'
                if pn.ntype == cn.ntype and pn.text_hash == cn.text_hash and cid not in curr_found:
                    prev_moved.append((pet, pidx))
                    curr_moved.append((cet, cidx))
                    curr_found.add(cid)
                    if pet == 'change':
                        corresponding_changed_node = diff['change'][pidx][1]
                        add_to_insert[cidx] = corresponding_changed_node
                    if cet == 'change':
                        corresponding_changed_node = diff['change'][cidx][0]
                        add_to_remove[pidx] = corresponding_changed_node
                    break

        # populate move list
        # if from a change, make sure it also isn't set to be moved to insert/remove
        diff['move'] = []
        if prev_moved:
            for i in range(len(prev_moved)):
                pet, pidx = prev_moved[i]
                cet, cidx = curr_moved[i]
                pn = diff[pet][pidx]
                if pet == 'change':  # pn is not the node but tuple of (prev_node, curr_node)
                    pn = pn[0]
                    if pidx in add_to_remove:  # don't add to remove -- was involved in its own move
                        add_to_remove.pop(pidx)
                cn = diff[cet][cidx]
                if cet == 'change':
                    cn = cn[1]
                    if cidx in add_to_insert:
                        add_to_insert.pop(cidx)
                diff['move'].append((pn, cn))
            prev_moved = sorted(prev_moved, reverse=True)
            for pet, pidx in prev_moved:
                diff[pet].pop(pidx)
            curr_moved = sorted(curr_moved, reverse=True)
            for cet, cidx in curr_moved:
                if (cet, cidx) not in prev_moved:  # was part of a change, already popped
                    diff[cet].pop(cidx)

            diff['insert'].extend(list(add_to_insert.values()))
            diff['remove'].extend(list(add_to_remove.values()))


class Diff:
    """
    Diff result with helper functions for post-processing / cleaning up the result
    """

    def __init__(self, nodes_removed, nodes_inserted, nodes_changed, nodes_moved, debug=False):
        self.remove = [n.dump(debug) for n in nodes_removed]
        self.insert = [n.dump(debug) for n in nodes_inserted]
        self.change = [{'prev': pn.dump(debug), 'curr': cn.dump(debug)} for pn, cn in nodes_changed]
        self.move = [{'prev': pn.dump(debug), 'curr': cn.dump(debug)} for pn, cn in nodes_moved]
        self.sections_p_to_c = {}
        self.sections_c_to_p = {}
        self.processed = False
        self.debug = debug

    def post_process(self, sections_prev, sections_curr, lang):
        if not self.processed:
            self._section_mapping(sections_prev, sections_curr)
            self._merge_text_changes(sections_prev, sections_curr, lang)
            self._build_result(sections_prev, sections_curr)
            self.processed = True
        return self.result

    def _build_result(self, sections_prev, sections_curr):
        sp = {}
        sc = {}
        for n in self.remove:
            sec_name = n['section']
            sp[sec_name] = sections_prev[sec_name]
        for n in self.insert:
            sec_name = n['section']
            # update name to section in previous revision for consistency (if it exists)
            sec_id = self.sections_c_to_p[sec_name] or sec_name
            n['section'] = sec_id
            sc[sec_id] = sections_curr[sec_name]
        for n in self.change + self.move:
            pn = n['prev']
            cn = n['curr']
            psec_name = pn['section']
            sp[psec_name] = sections_prev[psec_name]
            csec_name = cn['section']
            # update name to section in previous revision for consistency (if it exists)
            csec_id = self.sections_c_to_p[csec_name] or csec_name
            cn['section'] = csec_id
            sc[csec_id] = sections_curr[csec_name]

        self.result = {'remove': self.remove, 'insert': self.insert, 'change': self.change, 'move': self.move}
        if self.debug:
            self.result['sections-prev'] = sp
            self.result['sections-curr'] = sc

    def _section_mapping(self, sections_prev, sections_curr):
        """Build mapping of sections between previous and current versions of article."""
        prev = list(sections_prev.keys())
        curr = list(sections_curr.keys())
        p_to_c = {}
        c_to_p = {}
        removed = []
        # removed sections map to null in current; inserted sections map to null in previous
        for n in self.remove:
            if n['type'] == 'Heading':
                for i, s in enumerate(prev):
                    if s == n['section']:
                        removed.append(i)
                        break
        for idx in sorted(removed, reverse=True):
            p_to_c[prev[idx]] = None
            prev.pop(idx)
        inserted = []
        for n in self.insert:
            if n['type'] == 'Heading':
                for i, s in enumerate(curr):
                    if s == n['section']:
                        inserted.append(i)
                        break
        for idx in sorted(inserted, reverse=True):
            c_to_p[curr[idx]] = None
            curr.pop(idx)

        # changes happen in place so don't effect structure of doc and can be ignored
        # for moved sections, reorder mapping so they are aligned again for dumping
        for c in self.move:
            pn = c['prev']
            cn = c['curr']
            if pn['type'] == 'Heading':
                prev_idx = None
                curr_idx = None
                for i, s in enumerate(prev):
                    if s == pn['section']:
                        prev_idx = i
                        break
                for i, s in enumerate(curr):
                    if s == cn['section']:
                        curr_idx = i
                        break
                if prev_idx is not None and curr_idx is not None:
                    s = curr.pop(curr_idx)
                    curr.insert(prev_idx, s)

        for i in range(len(prev)):
            p_to_c[prev[i]] = curr[i]
            c_to_p[curr[i]] = prev[i]

        self.sections_p_to_c = p_to_c
        self.sections_c_to_p = c_to_p

    def _merge_text_changes(self, sections_prev, sections_curr, lang='en'):
        """Replace isolated text changes with section-level text changes."""
        changes = []
        # check each previous section and add if not a match with its corresponding current section
        for psec in self.sections_p_to_c:
            csec = self.sections_p_to_c[psec]
            if csec is None:
                prev_text = ''.join([extract_text(n, lang) for n in mw.parse(sections_prev[psec]).nodes])
                changes.append({'prev': {'name': node_to_name(prev_text), 'type': 'Text', 'text': prev_text,
                                         'section': psec, 'offset': 0}})
            elif sections_prev[psec] != sections_curr[csec]:
                prev_text = ''.join([extract_text(n, lang) for n in mw.parse(sections_prev[psec]).nodes])
                curr_text = ''.join([extract_text(n, lang) for n in mw.parse(sections_curr[csec]).nodes])
                if prev_text != curr_text:
                    changes.append({'prev': {'name': node_to_name(prev_text), 'type': 'Text', 'text': prev_text,
                                                 'section': psec, 'offset': 0},
                                        'curr': {'name': node_to_name(curr_text), 'type': 'Text', 'text': curr_text,
                                                 'section': csec, 'offset': 0}})
        # add in unmatched sections from current (new sections)
        for csec in self.sections_c_to_p:
            psec = self.sections_c_to_p[csec]
            if psec is None:
                curr_text = ''.join([extract_text(n, lang) for n in mw.parse(sections_curr[csec]).nodes])
                changes.append({'curr': {'name': node_to_name(curr_text), 'type': 'Text', 'text': curr_text,
                                         'section': csec, 'offset': 0}})

        # remove individual text changes
        for idx in range(len(self.remove) - 1, -1, -1):
            if self.remove[idx]['type'] == 'Text':
                self.remove.pop(idx)
        for idx in range(len(self.insert) - 1, -1, -1):
            if self.insert[idx]['type'] == 'Text':
                self.insert.pop(idx)
        for idx in range(len(self.change) - 1, -1, -1):
            if self.change[idx]['prev']['type'] == 'Text':
                self.change.pop(idx)

        # add in section-level text changes
        for c in changes:
            if 'prev' in c and 'curr' in c:
                self.change.append(c)
            elif 'prev' in c:
                self.remove.append(c['prev'])
            elif 'curr' in c:
                self.insert.append(c['curr'])



import re
import string
from collections import Counter
from mwedittypes.constants import *

#Universal regex for punctuations

class Tokenizer:

    def __init__(self, english_unicode, non_english_unicode, lang='en'):
        self.english_punc_regex = r'[{0}]'.format(re.escape(string.punctuation))
        self.english_unicode = english_unicode
        self.non_english_unicode =  non_english_unicode
        self.lang = lang

    def get_punctuations(self, text):
        #get ellipsis
        ellipsis = re.findall(r'\.{3,}',text)
        text = re.sub(r'\.{3,}','',text)

        #Get other punctuations
        english_punc_regex = re.findall(self.english_punc_regex, text)

        english_unicode_pattern = re.compile(self.english_unicode, re.UNICODE)
        english_unicode_regex = re.findall(english_unicode_pattern, text)

        non_english_unicode_pattern = re.compile(self.non_english_unicode, re.UNICODE)
        non_english_punc_regex = re.findall(non_english_unicode_pattern, text)

        return ellipsis + english_punc_regex +english_unicode_regex + non_english_punc_regex


    def get_whitespace(self, text):
        #Get whitespaces. Detects newlines, return characters as well as spaces as whitespaces.
        whitespace = re.findall(r'[\s]',text)
        return whitespace
    

    def get_words(self, text):
        #This extracts words inclusive of those with hyphens and apostrophes
        if self.lang in NON_WHITESPACE_LANGUAGES:
            word_list = re.findall(r"[\w]", text)

        else:
            # Get words with optionally hyphens and apostrophe:
            # \b - word breaks at start and end of word
            # \w+ - matches 1 or more alphanumeric characters
            # [-']? allows for hyphen/apostrophes within word
            # ((?:...)+) capturing group (extract words) made up of 1+ non-capturing sequences of
            # alphanumeric + hyphen/apostrophe. this allows for many-time-hyphenated words and the non-capturing wrapped
            # in capturing is to only gather the entire word (otherwise capturing groups only capture the last instance)
            word_list = re.findall(r"\b((?:\w+[-']?)+)\b", text)
        return word_list

    def get_sentences(self, text):
        #This regex accounts for ellipsis and splits accordingly
        #Minimum sentence size is three words. So a sentence needs to have atleast 3 words in it
        min_sentence_size = 2
        #Add non-english sentence breaks
        sentences = re.split(SENTENCE_BREAKS_REGEX, text)
        sentences = [s.strip() for s in sentences if len(self.get_words(s)) >= min_sentence_size]
        return sentences

    def get_paragraphs(self, text):
        if text != '':
            paragraphs = [paragraph.strip() for paragraph in re.split(r'\n{2}', text) if len(self.get_words(paragraph)) > 0]

            return paragraphs
        return []

    def tokenize_and_count(self, text):
        if self.lang in NON_WHITESPACE_LANGUAGES:
            word_key = 'Character'
        else:
            word_key = 'Word'

        return {'Whitespace':len(self.get_whitespace(text)),
                'Punctuation':len(self.get_punctuations(text)),
                word_key: len(self.get_words(text)),
                'Sentence':len(self.get_sentences(text)),
                'Paragraph': len(self.get_paragraphs(text))
                }

    def tokenize_and_get_occurence(self, text):
        whitespaces = self.get_whitespace(text)
        punctuation = self.get_punctuations(text)
        words = self.get_words(text)
        sentences = self.get_sentences(text)
        paragraphs = self.get_paragraphs(text)

        whitespace_occurence = Counter(whitespaces)
        punctuation_occurence = Counter(punctuation)
        words_occurence = Counter(words)
        sentences_occurence = Counter(sentences)
        paragraphs_occurence = Counter(paragraphs)

        if self.lang in NON_WHITESPACE_LANGUAGES:
            word_key = 'Character'
        else:
            word_key = 'Word'
        return {
            'Whitespace':dict(whitespace_occurence),
            'Punctuation':dict(punctuation_occurence),
            word_key:dict(words_occurence),
            'Sentence':dict(sentences_occurence),
            'Paragraph':dict(paragraphs_occurence)
        }


import re  # not technically needed because part of mwedittypes.tokenizer but that's confusing

import mwparserfromhell as mw
from mwedittypes.tokenizer import *
from mwedittypes.constants import *

def is_change_in_edit_type(node_type,prev_wikitext='',curr_wikitext=''):
    """ Checks if a change occurs in wikitexts

    Parameters
    ----------
    prev_wikitext : str
        Previous Wikitext
    curr_wikitext : str
        Current Wikitext
    node_type: str
        Node type
    Returns
    -------
    tuple
        Tuple containing the bool and edit type
    """
    try:
        prev_parsed_text = mw.parse(prev_wikitext)
        curr_parsed_text = mw.parse(curr_wikitext) 

        if node_type == 'Template':
            if prev_wikitext != '' and curr_wikitext != '':
                prev_temp_dict = { str(temp.name):str(temp.value) for temp in prev_parsed_text.filter_templates(recursive=False)[0].params}
                prev_temp_dict = dict(prev_temp_dict, **{'template_name':str(prev_parsed_text.filter_templates(recursive=False)[0].name)})
                
                curr_temp_dict = { str(temp.name):str(temp.value) for temp in curr_parsed_text.filter_templates(recursive=False)[0].params}
                curr_temp_dict = dict(curr_temp_dict, **{'template_name':str(curr_parsed_text.filter_templates(recursive=False)[0].name)})
                
                #Get the difference between template parameters. If it is more than 0, then a change occured
                if len(set(curr_temp_dict.items()) ^ set(prev_temp_dict.items())) > 0:
                    return True, 'Template'
            elif prev_wikitext != '' and curr_wikitext=='':
                templates = prev_parsed_text.filter_templates(recursive=False)
                if len(templates) > 0:
                    return True, 'Template'

            elif prev_wikitext == '' and curr_wikitext != '':
                templates = curr_parsed_text.filter_templates(recursive=False)
                if len(templates) > 0:
                    return True, 'Template'
            
        elif node_type == 'Media':
            if prev_wikitext != '' and curr_wikitext != '':
                if len(prev_parsed_text) > 0 and len(curr_parsed_text) > 0:
                    return True, 'Media'
            elif prev_wikitext != '' and curr_wikitext=='':
                if len(prev_parsed_text) > 0:
                    return True, 'Media'
            elif prev_wikitext == '' and curr_wikitext != '':
                if len(curr_parsed_text) > 0:
                    return True, 'Media'
        
        elif node_type == 'Category':
            if prev_wikitext != '' and curr_wikitext != '':
                prev_cat = prev_parsed_text.filter_wikilinks(recursive=False)
                curr_cat = curr_parsed_text.filter_wikilinks(recursive=False)
                if len(prev_cat) > 0 and len(curr_cat) > 0:
                    if prev_cat[0].text != curr_cat[0].text or \
                        prev_cat[0].title != curr_cat[0].title:
                        return True, 'Category'
                        
            elif prev_wikitext != '' and curr_wikitext=='':
                category = prev_parsed_text.filter_wikilinks(recursive=False)
                if len(category) > 0:
                    return True, 'Category'
            elif prev_wikitext == '' and curr_wikitext != '':
                category = curr_parsed_text.filter_wikilinks(recursive=False)
                if len(category) > 0:
                    return True, 'Category'

        elif node_type == 'Wikilink':
            if prev_wikitext != '' and curr_wikitext != '':
                prev_wikilink = prev_parsed_text.filter_wikilinks(recursive=False)
                curr_wikilink = curr_parsed_text.filter_wikilinks(recursive=False)

                if len(prev_wikilink) > 0 and len(curr_wikilink) > 0:
                    if prev_wikilink[0].text != curr_wikilink[0].text or \
                        prev_wikilink[0].title != curr_wikilink[0].title:
                        return True, 'Wikilink'
            elif prev_wikitext != '' and curr_wikitext=='':
                wikilinks = prev_parsed_text.filter_wikilinks(recursive=False)
                if len(wikilinks) > 0:
                    return True, 'Wikilink' 
            elif prev_wikitext == '' and curr_wikitext != '':
                wikilinks = curr_parsed_text.filter_wikilinks(recursive=False)
                if len(wikilinks) > 0:
                    return True, 'Wikilink'
        
        elif node_type == 'Reference':
            #Check if a reference changes
            if prev_wikitext != '' and curr_wikitext != '':
                prev_filtered_ref = prev_parsed_text.filter_tags(matches=lambda node: node.tag == "ref",recursive=False)
                curr_filtered_ref = curr_parsed_text.filter_tags(matches=lambda node: node.tag == "ref",recursive=False)
                if len(prev_filtered_ref) > 0 and len(curr_filtered_ref) > 0:
                    if prev_filtered_ref[0] != curr_filtered_ref[0]:
                        return True, 'Reference'
            elif prev_wikitext != '' and curr_wikitext=='':
                ref = prev_parsed_text.filter_tags(matches=lambda node: node.tag == "ref",recursive=False)
                if len(ref) > 0:
                    return True, 'Reference'

            elif prev_wikitext == '' and curr_wikitext != '':
                ref = curr_parsed_text.filter_tags(matches=lambda node: node.tag == "ref",recursive=False)
                if len(ref) > 0:
                    return True, 'Reference'

        elif node_type == 'Table':
            if prev_wikitext != '' and curr_wikitext != '':
                #Check if a table changes
                prev_filtered_table = prev_parsed_text.filter_tags(matches=lambda node: node.tag == "table",recursive=False)
                curr_filtered_table = curr_parsed_text.filter_tags(matches=lambda node: node.tag == "table",recursive=False)

                if len(prev_filtered_table) > 0 and len(curr_filtered_table) > 0:
                    if prev_filtered_table[0].contents != curr_filtered_table[0].contents:
                        return True, 'Table'

            elif prev_wikitext != '' and curr_wikitext=='':
                tables = prev_parsed_text.filter_tags(matches=lambda node: node.tag == "table",recursive=False)
                if len(tables) > 0:
                    return True, 'Table'

            elif prev_wikitext == '' and curr_wikitext != '':
                tables = curr_parsed_text.filter_tags(matches=lambda node: node.tag == "table",recursive=False)
                if len(tables) > 0:
                    return True, 'Table'

        elif node_type == 'Text Formatting':
            #Check if a text format changes
            if prev_wikitext != '' and curr_wikitext != '':
                prev_filtered_text_formatting = prev_parsed_text.filter_tags(matches=lambda node: node.tag in TEXT_FORMATTING_TAGS,recursive=False)
                curr_filtered_text_formatting = curr_parsed_text.filter_tags(matches=lambda node: node.tag in TEXT_FORMATTING_TAGS,recursive=False)

                if len(prev_filtered_text_formatting) > 0 and len(curr_filtered_text_formatting) > 0:
                    if prev_filtered_text_formatting[0].tag != curr_filtered_text_formatting[0].tag:
                        return True, 'Text Formatting'
            elif prev_wikitext != '' and curr_wikitext=='':
                text_format = prev_parsed_text.filter_tags(matches=lambda node: node.tag in TEXT_FORMATTING_TAGS,recursive=False)
                if len(text_format) > 0:
                    return True, 'Text Formatting'
            elif prev_wikitext == '' and curr_wikitext != '':
                text_format = curr_parsed_text.filter_tags(matches=lambda node: node.tag in TEXT_FORMATTING_TAGS,recursive=False)
                if len(text_format) > 0:
                    return True, 'Text Formatting'

        elif node_type == 'List':
            if prev_wikitext != '' and curr_wikitext != '':
                #Check if a list changes
                prev_filtered_list = prev_parsed_text.filter_tags(matches=lambda node: node.tag in LIST_TAGS,recursive=False)
                curr_filtered_list = curr_parsed_text.filter_tags(matches=lambda node: node.tag in LIST_TAGS,recursive=False)

                if len(prev_filtered_list) > 0  and len(curr_filtered_list) > 0:
                    if prev_filtered_list[0].contents != curr_filtered_list[0].contents:
                        return True, 'List'

            elif prev_wikitext != '' and curr_wikitext=='':
                lists = prev_parsed_text.filter_tags(matches=lambda node: node.tag in LIST_TAGS,recursive=False)
                if len(lists) > 0:
                    return True, 'List'

            elif prev_wikitext == '' and curr_wikitext != '':
                lists = curr_parsed_text.filter_tags(matches=lambda node: node.tag in LIST_TAGS,recursive=False)
                if len(lists) > 0:
                    return True, 'List'


        elif node_type == 'Table Element':
            if prev_wikitext != '' and curr_wikitext != '':
                #Check if a table element changes
                prev_filtered_table = prev_parsed_text.filter_tags(matches=lambda node: node.tag in TABLE_ELEMENTS_TAGS,recursive=False)
                curr_filtered_table = curr_parsed_text.filter_tags(matches=lambda node: node.tag in TABLE_ELEMENTS_TAGS,recursive=False)

                if len(prev_filtered_table) > 0 and len(curr_filtered_table) > 0:
                    if prev_filtered_table[0].contents != curr_filtered_table[0].contents:
                        return True, 'Table Element'

            elif prev_wikitext != '' and curr_wikitext=='':
                table_elems = prev_parsed_text.filter_tags(matches=lambda node: node.tag in TABLE_ELEMENTS_TAGS,recursive=False)
                if len(table_elems) > 0:
                    return True, 'Table Element'

            elif prev_wikitext == '' and curr_wikitext != '':
                table_elems = curr_parsed_text.filter_tags(matches=lambda node: node.tag in TABLE_ELEMENTS_TAGS,recursive=False)
                if len(table_elems) > 0:
                    return True, 'Table Element'

        elif node_type == 'Gallery':
            #Check if a gallery changes
            if prev_wikitext != '' and curr_wikitext != '':
                prev_filtered_gallery = prev_parsed_text.filter_tags(matches=lambda node: node.tag == 'gallery',recursive=False)
                curr_filtered_gallery = curr_parsed_text.filter_tags(matches=lambda node: node.tag == 'gallery',recursive=False)

                if len(prev_filtered_gallery) > 0  and len(curr_filtered_gallery) > 0:
                    if prev_filtered_gallery[0].contents != curr_filtered_gallery[0].contents:
                        return True, 'Gallery'

            elif prev_wikitext != '' and curr_wikitext=='':
                gallery = prev_parsed_text.filter_tags(matches=lambda node: node.tag == 'gallery',recursive=False)
                if len(gallery) > 0:
                    return True, 'Gallery'

            elif prev_wikitext == '' and curr_wikitext != '':
                gallery = curr_parsed_text.filter_tags(matches=lambda node: node.tag == 'gallery',recursive=False)
                if len(gallery) > 0:
                    return True, 'Gallery'
            
        elif 'Tag' in node_type:
            #Check if a tag changes
            if prev_wikitext != '' and curr_wikitext != '':
                prev_filtered_tag = prev_parsed_text.filter_tags(recursive=False)
                curr_filtered_tag = curr_parsed_text.filter_tags(recursive=False)

                if len(prev_filtered_tag) > 0  and len(curr_filtered_tag) > 0:
                    if prev_filtered_tag[0].contents != curr_filtered_tag[0].contents:
                        return True, node_type

            elif prev_wikitext != '' and curr_wikitext=='':
                tags = prev_parsed_text.filter_tags(recursive=False)
                if len(tags) > 0:
                    return True, node_type

            elif prev_wikitext == '' and curr_wikitext != '':
                tags = curr_parsed_text.filter_tags(recursive=False)
                if len(tags) > 0:
                    return True, node_type

        elif node_type == 'Heading':
            if prev_wikitext != '' and curr_wikitext != '':
                prev_filtered_section = prev_parsed_text.filter_headings(recursive=False)[0]
                curr_filtered_section = curr_parsed_text.filter_headings(recursive=False)[0]

                if prev_filtered_section.title != curr_filtered_section.title:
                    return True, 'Heading'
            elif prev_wikitext != '' and curr_wikitext=='':
                section = prev_parsed_text.filter_headings(recursive=False)
                if len(section) > 0:
                    return True, 'Heading'
            elif prev_wikitext == '' and curr_wikitext != '':
                section = curr_parsed_text.filter_headings(recursive=False)
                if len(section) > 0:
                    return True, 'Heading'

        elif node_type == 'Comment':
            if prev_wikitext != '' and curr_wikitext != '':
                prev_filtered_comments = prev_parsed_text.filter_comments(recursive=False)[0]
                curr_filtered_comments = curr_parsed_text.filter_comments(recursive=False)[0]

                if prev_filtered_comments.contents != curr_filtered_comments.contents:
                    return True, 'Comment'

            elif prev_wikitext != '' and curr_wikitext=='':
                comments = prev_parsed_text.filter_comments(recursive=False)
                if len(comments) > 0:
                    return True, 'Comment'

            elif prev_wikitext == '' and curr_wikitext != '':
                comments = curr_parsed_text.filter_comments(recursive=False)
                if len(comments) > 0:
                    return True, 'Comment'

        elif node_type == 'External Link':
            if prev_wikitext != '' and curr_wikitext != '':
                prev_filtered_external_links = prev_parsed_text.filter_external_links(recursive=False)[0]
                curr_filtered_external_links = curr_parsed_text.filter_external_links(recursive=False)[0]

                if prev_filtered_external_links.title != curr_filtered_external_links.title or \
                    prev_filtered_external_links.url != curr_filtered_external_links.url:
                    return True, 'External Link'
            elif prev_wikitext != '' and curr_wikitext=='':
                external_links = prev_parsed_text.filter_external_links(recursive=False)
                if len(external_links) > 0:
                    return True, 'External Link'

            elif prev_wikitext == '' and curr_wikitext != '':
                external_links = curr_parsed_text.filter_external_links(recursive=False)
                if len(external_links) > 0:
                    return True, 'External Link'
        else:
            return True, node_type
    except Exception:
        pass
    return False, None


def parse_change_text(node_type, prev_wikitext='',curr_wikitext='', lang='en'):
    # Initialize tokenizer class
    tokenizer = Tokenizer(ENGLISH_UNICODE, NON_ENGLISH_UNICODE, lang=lang)

    if node_type == 'Text':
        if prev_wikitext == '' and curr_wikitext != '':
            edittype = 'insert'
        elif prev_wikitext != '' and curr_wikitext == '':
            edittype = 'remove'
        elif prev_wikitext != '' and curr_wikitext != '':
            edittype = 'change'
        else:
            edittype = None

        
        if edittype:
            prev_tokenizer = tokenizer.tokenize_and_get_occurence(prev_wikitext)
            curr_tokenizer = tokenizer.tokenize_and_get_occurence(curr_wikitext)

            result = {}
            for text_category in curr_tokenizer.keys():
                items_diff_list = list(set(curr_tokenizer[text_category].items())  ^ set(prev_tokenizer[text_category].items()))
                for item in items_diff_list:
                    diff = curr_tokenizer[text_category].get(item[0], 0) - prev_tokenizer[text_category].get(item[0],0)
                    result[text_category] = dict(result.get(text_category,{}), **{item[0]:diff})

            #Get the maximum value between the sum of positives and sum of negatives
                if len(result.get(text_category,{})) > 0: 
                    removals = sum(abs(item) for item in result[text_category].values() if item < 0)
                    additions = sum(abs(item) for item in result[text_category].values() if item > 0)
                    change = min(removals, additions)
                    result[text_category] = {}
                    removals -= change
                    additions -= change

                    if removals > 0:
                        result[text_category]['remove'] = removals
                    
                    if additions > 0:
                        result[text_category]['insert'] = additions

                    if change > 0:
                        result[text_category]['change'] = change
            return result

        return None

def get_diff_count(result, lang='en'):
    """ Gets the edit type count of a diff

    Parameters
    ----------
    result : dict
        The diff API response containing inserts,removes and changes made in a Wikipedia revision.
    Returns
    -------
    dict
        a dict containing a count of edit type occurence
    """

    edit_types = {}
    section_titles = set()
    prev_text = []
    curr_text = []
    # loop through all removed nodes
    for r in result['remove']:
        text = r['text']  # wikitext of the node
        # if node is text, just check whether there's anything and retain for later
        # because all the text is processed at once at the end
        if r['type'] == 'Text' and text:
            prev_text.append(text)
            section_titles.add(r['section'])
        # non-text node: verify/fine-tune the edit type and add to results dictionary
        else:
            is_edit_type_found,edit_type = is_change_in_edit_type(r['type'],text,'')
            if is_edit_type_found:
                section_titles.add(r['section'])
                if edit_types.get(edit_type,{}):
                    edit_types[edit_type]['remove'] = edit_types[edit_type].get('remove', 0) + 1
                else:
                    edit_types[edit_type] = {'remove':1}

    for i in result['insert']:
        text = i['text']
        if i['type'] == 'Text' and text:
            curr_text.append(text)
            section_titles.add(i['section'])
        else:
            is_edit_type_found,edit_type = is_change_in_edit_type(i['type'],'',text)
            #check if edit_type in edit types dictionary
            if is_edit_type_found:
                section_titles.add(i['section'])
                if edit_types.get(edit_type,{}):
                    edit_types[edit_type]['insert'] = edit_types[edit_type].get('insert', 0) + 1
                else:
                    edit_types[edit_type] = {'insert':1}

    for c in result['change']:
        if c['prev']['type'] == c['curr']['type']:
            if c['prev']['type'] == 'Text' and c['prev']['text'] != c['curr']['text']:
                prev_text.append(c['prev']['text'])
                curr_text.append(c['curr']['text'])
                section_titles.add(c['curr']['section'])
                section_titles.add(c['prev']['section'])
            else:
                is_edit_type_found,edit_type = is_change_in_edit_type(c['prev']['type'],c['prev']['text'],c['curr']['text'])
                #check if edit_type in edit types dictionary
                if is_edit_type_found:
                    section_titles.add(c['curr']['section'])
                    section_titles.add(c['prev']['section'])
                    if edit_types.get(edit_type,{}):
                        edit_types[edit_type]['change'] = edit_types[edit_type].get('change', 0) + 1
                    else:
                        edit_types[edit_type] = {'change':1}

    for m in result['move']:
        text = m['prev']['text']
        is_edit_type_found,edit_type = is_change_in_edit_type(m['prev']['type'],text,'')
        #check if edit_type in edit types dictionary
        if is_edit_type_found:
            section_titles.add(m['curr']['section'])
            section_titles.add(m['prev']['section'])
            if edit_types.get(edit_type,{}):
                edit_types[edit_type]['move'] = edit_types[edit_type].get('move', 0) + 1
            else:
                edit_types[edit_type] = {'move':1}

    # join together all the changed section text and process
    if prev_text or curr_text:
        is_text_change_found = parse_change_text('Text', ''.join(prev_text), ''.join(curr_text), lang=lang)
        if is_text_change_found:
            for text_subcat,text_et in is_text_change_found.items():
                edit_types[text_subcat] = {}
                for et, et_count in text_et.items():
                    edit_types[text_subcat][et] = edit_types[text_subcat].get(et, 0) + et_count
    
    if section_titles:
        if 'Section' not in edit_types:
            edit_types['Section'] = {}
        sec_remove = edit_types.get('Heading', {}).get('remove', 0)
        sec_insert = edit_types.get('Heading', {}).get('insert', 0) + int('prev-no-content' in result)
        sec_change = len(section_titles) - sec_remove - sec_insert
        if sec_remove:
            edit_types['Section']['remove'] = sec_remove
        if sec_insert:
            edit_types['Section']['insert'] = sec_insert
        if sec_change:
            edit_types['Section']['change'] = sec_change
    return edit_types


from .mwedittypes import EditTypes

__title__ = "mwedittypes"
__summary__ = "mwedittypes is a package that supports edit diffs and action detection for Wikipedia"
__url__ = "https://github.com/geohci/edit-types"

__version__ = "1.1.5"

__license__ = "MIT License"

__all__ = ["EditTypes"]



from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

VERSION = '1.1.5'
DESCRIPTION = 'Edit diffs and type detection for Wikipedia'
LONG_DESCRIPTION = 'A package that allows edit diffs and type detection for Wikipedia.'

# Dev dependencies
EXTRAS_REQUIRE = {
    "tests": ["pytest>=6.2.5"],
}

EXTRAS_REQUIRE["dev"] = (
    EXTRAS_REQUIRE["tests"]
)

# Setting up
setup(
    name="mwedittypes",
    version=VERSION,
    author="geohci & Amamgbu (Isaac Johnson & Jesse Amamgbu)",
    author_email="<amamgbujesse@yahoo.com>",
    url="https://github.com/geohci/edit-types",
    license="MIT License",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=['mwparserfromhell', 'anytree'],
    keywords=['python', 'wikipedia', 'edit types', 'edit diffs', 'wiki', 'edit detection'],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        
    ],
    extras_require=EXTRAS_REQUIRE,
    include_package_data=True,
    zip_safe=False,
)